@@ -50,21 +50,7 @@
   :type 'string
   :group 'ekg-agent)
 
-(defcustom ekg-agent-self-instruct-tag "agent/instruct"
-  "The tag used to identify notes with instructions for the agent."
-  :type 'string
-  :group 'ekg-agent)
-
 (add-to-list 'ekg-hidden-tags ekg-agent-self-info-tag)
-(add-to-list 'ekg-hidden-tags ekg-agent-self-instruct-tag)
-
-(defcustom ekg-agent-context-func #'ekg-agent-starting-context
-  "Function that returns the starting information for an ekg agent.
-
-This is combined with a standard prompt to form the LLM context for the
-agent."
-  :type 'function
-  :group 'ekg-agent)
 
 (defcustom ekg-agent-daily-time "09:00"
   "Time of day to run the daily agent evaluation, in HH:MM format (24-hour).
@@ -126,6 +112,17 @@ stdout will be returned."
 (defvar ekg-agent--daily-timer nil
   "Timer object for the daily agent evaluation.")
 
+(defun ekg-agent--provider ()
+  "Return the provider for the LLM.
+
+If there is a list of providers, find the first one that has tool
+calling, or if none have tool calling, just return the first provider."
+  (if (listp ekg-llm-provider)
+      (or (car (seq-filter (lambda (provider) (member 'tool-use (llm-capabilities provider)))
+                           ekg-llm-provider))
+          (car ekg-llm-provider))
+    ekg-llm-provider))
+
 (defmacro ekg-agent--with-error-as-text (&rest body)
   "Execute BODY, returning any error as a descriptive string.
 If BODY signals an error, return \"Error: MESSAGE\" instead of
@@ -374,7 +371,7 @@ EXTRA-TOOLS is a list of additional tools beyond `ekg-agent-base-tools' and `ekg
    (append ekg-agent-base-tools
            ekg-agent-extra-tools
            extra-tools
-           (when (llm-google-p (ekg-llm--provider))
+           (when (llm-google-p (ekg-agent--provider))
              (list (make-llm-tool :function #'ignore
                                   :name "google_search"
                                   :description "Google Search built-in tool"
@@ -483,10 +480,9 @@ EXTRA-TOOLS is a list of additional tools beyond `ekg-agent-base-tools' and `ekg
   "Return a message instructing the agent to save state before stopping."
   (format (concat
            "Time limit reached. Before this session ends, immediately create a note with your current state "
-           "using `create_note` (tag it with `%s` or `%s`). Also call `summarize_state` with a brief update. "
+           "using `create_note` (tag it with `%s`). Also call `summarize_state` with a brief update. "
            "Then finish by calling `end` or the appropriate completion tool.")
-          ekg-agent-self-info-tag
-          ekg-agent-self-instruct-tag))
+          ekg-agent-self-info-tag))
 
 (defun ekg-agent--summarize-state (state)
   "Write STATE to the agent log window."
@@ -845,11 +841,11 @@ agent."
                            major-mode)
                    (buffer-substring-no-properties start end))))
 
-(defun ekg-agent-starting-context ()
+(defun ekg-agent-latest-notes-context ()
   "Return the context for an agent for new sessions.
 
 This includes the latest 10 notes and the org agenda, and the last 10
-self info and self-instruct notes."
+self info."
   (concat
    "The last 10 notes:\n"
    (let ((ekg-llm-note-numwords 300))
@@ -860,9 +856,9 @@ self info and self-instruct notes."
       "\n\nTODO items from org-mode:\n"
       (with-temp-buffer (org-ql-search org-agenda-files '(todo) :buffer (current-buffer))
                         (buffer-substring-no-properties (point-min) (point-max)))))
-   "\n\nThe last 10 self-info and self-instruct notes:\n"
+   "\n\nThe last 10 self-info notes:\n"
    (mapconcat #'ekg-llm-note-to-text
-              (seq-take (ekg-get-notes-with-any-tags (list ekg-agent-self-info-tag ekg-agent-self-instruct-tag))
+              (seq-take (ekg-get-notes-with-tag ekg-agent-self-info-tag)
                         10)
               "\n\n")))
 
@@ -891,20 +887,75 @@ Returns nil if no AGENTS.md files are found."
       (concat "\n\nUser instructions from AGENTS.md files:\n\n"
               (string-join (nreverse parts) "\n\n")))))
 
+(defun ekg-agent-cotagged-prompt-tags ()
+  "Return a list of all tags that are cotagged with the prompt tag."
+  (let (result)
+    (dolist (note (ekg-get-notes-with-tag ekg-llm-prompt-tag))
+      (dolist (tag (ekg-note-tags note))
+        (when (not (string-equal tag ekg-llm-prompt-tag))
+          (push tag result))))
+    result))
+
 (defun ekg-agent-instructions-intro ()
-  "Introductory instructions for the ekg agent."
+  "Introductory instructions for the ekg agent.
+
+These instructions guide the agent on how to effectively use ekg
+notes as memory for tasks.  Follow these rules to ensure proper
+note-taking and knowledge retention."
   (let ((timeout-desc (if (and (numberp ekg-agent-timeout-seconds)
                                (> ekg-agent-timeout-seconds 0))
                           (format "%s seconds" ekg-agent-timeout-seconds)
                         "no timeout configured")))
-    (concat
-     (format
-      "You are an agent that works with a user through a note taking system in
-Emacs, called ekg. In ekg, notes have tags and content, and can be
-written in markdown, org-mode or plain text. Tags can be any string, and
-can have whitespace. They can be nested as well, such as
-`self improvement/languages/spanish/conjucations`. ekg is backed by a
-database, and the way to interact with it is with the tools provided.
+    (format
+     "IMPORTANT: Read this whole section before starting any task. These rules are mandatory, not optional.
+
+ekg is an external memory system that stores notes in a database
+organized by tags. When you work on a task, you **must** use ekg to:
+- **Remember** important context, decisions, and findings
+- **Share** knowledge across tasks and sessions
+- **Learn** from past work and avoid repeating mistakes
+
+The rules below ensure you use ekg effectively. Failure to follow these
+rules will result in lost context and repeated work.
+
+== MANDATORY WORKFLOW ==
+
+Before you begin any substantive work on a task:
+
+1. **INITIAL INVESTIGATION (DO NOT SKIP)**
+   - Look up the current project's ekg tags. If there is an EKG-PROJECT
+     property in the org file, use that as the project tag.  Or, look at
+     existing tags that match the description.
+   - Search the ekg database for existing notes with relevant tags.
+   - Read existing notes to understand what's already known.
+   - Read the project's AGENTS.md file (if present) or check
+     ~/.pi/agent/AGENTS.md for global instructions.
+   - Read existing notes that contain instructions for directories,
+     skills, or projects mentioned in the task description.  Each of
+     these will have a tag cotagged with the prompt tag.  Existing
+     cotags with the prompt tag are: %s.
+   - If you find relevant notes, reference them in your task description.
+
+2. **CREATE INITIAL TASK NOTE**
+   - Create a new note with:
+     - A descriptive title based on the task.
+     - Tags: at minimum `agent-task`, plus any project-specific tags.
+     - Include the org task ID and description as the core content.
+     - Note the ekg note ID returned (you'll need it for updates).
+   - This note is your \"task journal\" - all findings go here.  It
+     will be the source of information if we need to recover the state
+     of the work in progress.
+
+3. **DOCUMENT PROGRESS THROUGHOUT THE TASK**
+   - Before you start significant work: add an entry describing what you plan to do.
+   - When you discover important information: add a note about it.
+   - When you hit a problem or make a decision: document the rationale.
+   - Be specific: include file paths, code snippets, key insights, and context.
+
+4. **CREATE NEW NOTES FOR GENERAL KNOWLEDGE**
+   - If you discover something that will help with other tasks (not just this one): create a new note.
+   - Tag it appropriately (e.g., `skill/<new skill>`, `<project>/lessons`).
+   - Link back to the task note if relevant.
 
 In org-mode, you can link to a note with `[[ekg-note:<id>][<link display
 text>]]`.  Tags can be linked with `[[ekg-tag:<tag>][<link display
@@ -913,29 +964,31 @@ text>]]`.
 In markdown mode, there's no way to link directly to notes, but you can
 use [[tag]] to link to a tag.
 
-These notes should act as memory for subjects and tasks that both users
-and agents can read and write. When you have learned something, you
-*must* write a note about it. Tag the note with any project id you have
-to represent the task you are working on, as a nested project struct.
-For example, the tags might be `[ekg/latency/agent async startup]`.
-Similarly, when you need to know something, look at the notes before
-asking the user. Always look at the notes before starting a task to
-understand the background.
+== FINAL STEPS (DO NOT SKIP) ==
 
-To write a note to your future self so that you remember important
-information the next time you are run, use the `%s` tag for information
-for your future self, so that you can behave more usefully in the future.
-Write notes with these tags if you feel you have discovered something
-that will make future runs better, and want to record this.
+When the task is complete:
+
+5. **RETROSPECTIVE AND KNOWLEDGE EXTRACTION**
+   - Ask: what did I learn that's generally useful?
+   - Convert those learnings into standalone notes with appropriate tags
+   - If something went wrong, document the lesson
+   - If my instructions could be improved, create or edit notes tagged
+     with `%s` and other tags that will allow you to later use it in the
+     correct contexts.
+
+6. **UPDATE THE TASK NOTE**
+   - Add a final entry summarizing the outcome
+   - Link to any new notes you created
+   - Mark the task as done in the org file
 
 Use the `summarize_state` tool regularly to write brief status updates
 to the user-visible agent log window (at least every couple of tool
 calls or whenever your plan changes).
 
-The session may end if a total task timeout has been configured. Or,
+The session may end if a total task timeout (%s) has been configured. Or,
 some error may interrupt the processing. Because your processing can end
 unexpected, you MUST occasionally write out your current state to an ekg
-note (tag with `%s` or `%s`, and a tag you choose to represent the
+note (tag with `%s`, and a tag you choose to represent the
 task). If you receive a timeout warning, do this immediately and then
 finish. When executing a long-running task (more than a couple of tool
 calls), start saving your state every few tool calls to a note.
@@ -943,11 +996,14 @@ calls), start saving your state every few tool calls to a note.
 When creating a note, text that you add will automatically have the tags
 surrounding it to indicate that it was written by an LLM.  Do not add
 these tags manually."
-      ekg-agent-self-instruct-tag
-      timeout-desc
-      ekg-agent-self-info-tag
-      ekg-agent-self-instruct-tag)
-     (or (ekg-agent--agents-md-context) ""))))
+     (concat "[" (string-join (ekg-agent-get-all-cotagged-prompt-tags) " ") "]")
+     ekg-llm-prompt-tag
+     timeout-desc
+     ekg-agent-self-info-tag)))
+
+(defun ekg-agent--load-org-instructions ()
+  "Legacy function - kept for backwards compatibility.
+This function now returns the default instructions as a fallback.")
 
 (defun ekg-agent-instructions-evaluate-status ()
   "Return instructions for the agent to evaluate user status and help them."
@@ -1005,7 +1061,7 @@ to create new notes or perform other actions to help the user."
 (defun ekg-agent--prompt-id (prompt)
   "From llm PROMPT, call an LLM to get a short identifier."
   (let (id)
-    (llm-chat (ekg-llm--provider)
+    (llm-chat (ekg-agent--provider)
               (llm-make-chat-prompt
                (llm-chat-prompt-interaction-content
                 (car (seq-filter (lambda (interaction)
@@ -1099,7 +1155,7 @@ session.  At iteration 0 the log buffer is created and
         (ekg-agent--prompt-append-user-message prompt (ekg-agent--timeout-warning-message))
         (setq timeout-final t))
       (llm-chat-async
-       (ekg-llm--provider)
+       (ekg-agent--provider)
        prompt
        (lambda (result)
          (if (and (buffer-live-p log-buf)
@@ -1279,8 +1335,7 @@ ARG, if non-nil, allows editing the instructions."
                                                 (ekg-get-notes-with-any-tags
                                                  (append
                                                   (ekg-note-tags ekg-note)
-                                                  (list ekg-agent-self-info-tag
-                                                        ekg-agent-self-instruct-tag))))
+                                                  (list ekg-agent-self-info-tag))))
                                     10))
            (context-notes-json (let ((ekg-llm-note-numwords 100))
                                  (mapconcat #'ekg-llm-note-to-text context-notes "\n\n")))
@@ -1323,7 +1378,7 @@ currently editing.\n\n"
 (defun ekg-agent-show-internal-notes ()
   "Show notes with agent internal tags."
   (interactive)
-  (ekg-show-notes-with-any-tags (list ekg-agent-self-info-tag ekg-agent-self-instruct-tag)))
+  (ekg-show-notes-with-tag ekg-agent-self-info-tag))
 
 ;;;###autoload
 (defun ekg-agent-add-note (text tags mode)
